from abc import (
    ABC,
    abstractmethod,
)
from asyncio import gather
from contextlib import contextmanager
from inspect import isawaitable
from typing import (
    List,
    Dict,
    Any,
    Optional,
    overload,
    Iterator,
    Union,
    Type,
)

from hiku.federation.engine import Engine
from hiku.federation.utils import representation_to_ident
from hiku.federation.introspection import (
    BaseFederatedGraphQLIntrospection,
    FederatedGraphQLIntrospection,
    AsyncFederatedGraphQLIntrospection,
    is_introspection_query,
    extend_with_federation,
)
from hiku.federation.validate import validate

from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.federation.denormalize import DenormalizeEntityGraphQL
from hiku.endpoint.graphql import (
    _type_names,
    _switch_graph,
    GraphQLError,
    _StripQuery,
)
from hiku.graph import (
    Graph,
    apply,
)
from hiku.query import Node
from hiku.result import Proxy, Reference
from hiku.readers.graphql import Operation


def _process_query(graph: Graph, query: Node) -> Node:
    stripped_query = _StripQuery().visit(query)
    errors = validate(graph, stripped_query)
    if errors:
        raise GraphQLError(errors=errors)
    else:
        return stripped_query


def denormalize_entities(
    graph: Graph,
    query: Node,
    result: Proxy,
) -> List[Dict[str, Any]]:
    entities_link = query.fields_map["_entities"]
    node = entities_link.node
    representations = entities_link.options["representations"]

    entities = []
    for r in representations:
        typename = r["__typename"]
        ident = representation_to_ident(r)
        result.__ref__ = Reference(typename, ident)
        data = DenormalizeEntityGraphQL(graph, result, typename).process(node)
        entities.append(data)

    return entities


class BaseFederatedGraphEndpoint(ABC):
    query_graph: Graph
    mutation_graph: Optional[Graph]

    @property
    @abstractmethod
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        pass

    def __init__(
        self,
        engine: Engine,
        query_graph: Graph,
        mutation_graph: Optional[Graph] = None,
    ):
        self.engine = engine

        introspection = self.introspection_cls(query_graph, mutation_graph)
        self.query_graph = apply(query_graph, [introspection])
        if mutation_graph is not None:
            self.mutation_graph = apply(mutation_graph, [introspection])
        else:
            self.mutation_graph = None

    @contextmanager
    def context(self, op: Operation) -> Iterator[Dict]:
        yield {}

    def postprocess_result(
        self, result: Proxy, graph: Graph, op: Operation
    ) -> Dict:
        if "_service" in op.query.fields_map:
            return {"_service": {"sdl": result["sdl"]}}
        elif "_entities" in op.query.fields_map:
            return {"_entities": denormalize_entities(graph, op.query, result)}

        type_name = _type_names[op.type]

        data = DenormalizeGraphQL(graph, result, type_name).process(op.query)
        if is_introspection_query(op.query):
            extend_with_federation(graph, data)
        return data


class BaseSyncFederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    @abstractmethod
    def execute(self, graph: Graph, op: Operation, ctx: Dict) -> Dict:
        pass

    @abstractmethod
    def dispatch(self, data: Dict) -> Dict:
        pass


class BaseAsyncFederatedGraphQLEndpoint(BaseFederatedGraphEndpoint):
    @abstractmethod
    async def execute(self, graph: Graph, op: Operation, ctx: Dict) -> Dict:
        pass

    @abstractmethod
    async def dispatch(self, data: Dict) -> Dict:
        pass


class FederatedGraphQLEndpoint(BaseSyncFederatedGraphQLEndpoint):
    """Can execute either regular or federated queries.
    Handles following fields of federated query:
        - _service
        - _entities
    """

    @property
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        return FederatedGraphQLIntrospection

    def execute(self, graph: Graph, op: Operation, ctx: Optional[Dict]) -> Dict:
        stripped_query = _process_query(graph, op.query)
        if "_service" in stripped_query.fields_map:
            result = self.engine.execute_service(graph, self.mutation_graph)
        else:
            result = self.engine.execute(graph, stripped_query, ctx or {}, op)

        assert isinstance(result, Proxy)
        return self.postprocess_result(result, graph, op)

    def dispatch(self, data: Dict) -> Dict:
        try:
            graph, op = _switch_graph(
                data,
                self.query_graph,
                self.mutation_graph,
            )
            with self.context(op) as ctx:
                result = self.execute(graph, op, ctx)
            return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class AsyncFederatedGraphQLEndpoint(BaseAsyncFederatedGraphQLEndpoint):
    @property
    def introspection_cls(self) -> Type[BaseFederatedGraphQLIntrospection]:
        return AsyncFederatedGraphQLIntrospection

    async def execute(
        self, graph: Graph, op: Operation, ctx: Optional[Dict]
    ) -> Dict:
        stripped_query = _process_query(graph, op.query)
        if "_service" in stripped_query.fields_map:
            coro = self.engine.execute_service(graph, self.mutation_graph)
        else:
            coro = self.engine.execute(graph, stripped_query, ctx)

        assert isawaitable(coro)
        result = await coro
        return self.postprocess_result(result, graph, op)

    async def dispatch(self, data: Dict) -> Dict:
        try:
            graph, op = _switch_graph(
                data,
                self.query_graph,
                self.mutation_graph,
            )

            with self.context(op) as ctx:
                result = await self.execute(graph, op, ctx)
            return {"data": result}
        except GraphQLError as e:
            return {"errors": [{"message": e} for e in e.errors]}


class AsyncBatchFederatedGraphQLEndpoint(AsyncFederatedGraphQLEndpoint):
    @overload
    async def dispatch(self, data: List[Dict]) -> List[Dict]:
        ...

    @overload
    async def dispatch(self, data: Dict) -> Dict:
        ...

    async def dispatch(
        self, data: Union[Dict, List[Dict]]
    ) -> Union[Dict, List[Dict]]:
        if isinstance(data, list):
            return list(
                await gather(*(super().dispatch(item) for item in data))
            )

        return await super().dispatch(data)



from dataclasses import dataclass
from typing import Any, Optional, Type


@dataclass
class ScalarInfo:
    name: str
    typ: Any


def _process_scalar(
    cls: Type,
    *,
    name: Optional[str] = None,
) -> Type:
    cls.__scalar_info__ = ScalarInfo(
        name=name or cls.__name__,
        typ=cls,
    )

    return cls


def scalar(
    cls: Optional[Type] = None,
    *,
    name: Optional[str] = None,
) -> Any:
    def wrap(cls: Type) -> Type:
        return _process_scalar(
            cls,
            name=name,
        )

    if cls is None:
        return wrap

    return wrap(cls)



from itertools import chain
from contextlib import contextmanager
from collections import Counter
from typing import (
    List,
    Any,
    Optional,
    Union,
    Iterable,
)

from ..directives import Deprecated
from ..graph import (
    GraphVisitor,
    Root,
    Field,
    Node,
    Link,
    Option,
    Graph,
)
from ..graph import AbstractNode, AbstractField, AbstractLink, AbstractOption

from .errors import Errors
from ..types import GenericMeta, OptionalMeta


class GraphValidationError(TypeError):
    def __init__(self, errors: List[str]) -> None:
        self.errors = errors
        errors_list = "\n".join("- {}".format(e) for e in errors)
        super(GraphValidationError, self).__init__("\n" + errors_list)


class GraphValidator(GraphVisitor):
    class _NameFormatter(GraphVisitor):
        def visit_node(self, obj: Node) -> Optional[str]:
            return obj.name

        def visit_root(self, obj: Root) -> str:
            return "root"

        def visit_link(self, obj: Link) -> str:
            return ".{}".format(obj.name)

        def visit_field(self, obj: Field) -> str:
            return ".{}".format(obj.name)

        def visit_option(self, obj: Option) -> str:
            return ":{}".format(obj.name)

    _name_formatter = _NameFormatter()
    _graph_accept_types = (AbstractNode,)
    _node_accept_types = (AbstractField, AbstractLink)
    _link_accept_types = (AbstractOption,)
    _field_accept_types = (AbstractOption,)

    def __init__(self, items: List[Node]) -> None:
        self.items = items
        self.errors = Errors()
        self._ctx: List = []

    @classmethod
    def validate(cls, items: List[Node]) -> None:
        validator = cls(items)
        validator.visit_graph_items(items)
        if validator.errors.list:
            raise GraphValidationError(validator.errors.list)

    @contextmanager
    def push_ctx(self, obj: Union[Node, Link, Field]) -> Any:
        self._ctx.append(obj)
        try:
            yield
        finally:
            self._ctx.pop()

    @property
    def ctx(self) -> Union[Node, Link, Field]:
        return self._ctx[-1]

    def _get_duplicates(self, names: Iterable[str]) -> List[str]:
        counter = Counter(names)
        return [k for k, v in counter.items() if v > 1]

    def _format_names(self, names: List[str]) -> str:
        return ", ".join('"{}"'.format(name) for name in names)

    def _format_types(self, objects: List[Any]) -> str:
        return ", ".join(map(repr, set(type(obj) for obj in objects)))

    def _format_path(self, obj: Optional[Any] = None) -> str:
        path = self._ctx + ([obj] if obj is not None else [])
        return "".join(self._name_formatter.visit(i) for i in path)

    def _validate_deprecated_duplicates(self, obj: Union[Field, Link]) -> None:
        deprecated_count = sum(
            (1 for d in obj.directives if isinstance(d, Deprecated))
        )
        if deprecated_count > 1:
            self.errors.report(
                'Deprecated directive must be used only once for "{}", found {}'.format(  # noqa: E501
                    self._format_path(obj), deprecated_count
                )
            )

    def visit_option(self, obj: Option) -> None:
        if (
            isinstance(obj.type, GenericMeta)
            and not isinstance(obj.type, OptionalMeta)
            and obj.default is None
        ):
            self.errors.report(
                'Non-optional option "{}" must have a default value'.format(
                    self._format_path(obj)
                ),
            )

    def visit_field(self, obj: Field) -> None:
        invalid = [
            f
            for f in obj.options
            if not isinstance(f, self._field_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Invalid types provided as field "{}" options: {}'.format(
                    self._format_path(obj), self._format_types(invalid)
                )
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_field(obj)

        self._validate_deprecated_duplicates(obj)

    def visit_link(self, obj: Link) -> None:
        assert isinstance(self.ctx, Node)

        invalid = [
            f for f in obj.options if not isinstance(f, self._link_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Invalid types provided as link "{}" options: {}'.format(
                    self._format_path(obj), self._format_types(invalid)
                )
            )
            return

        with self.push_ctx(obj):
            super(GraphValidator, self).visit_link(obj)

        graph_nodes_map = {e.name for e in self.items if e.name is not None}
        if obj.node not in graph_nodes_map:
            self.errors.report(
                'Link "{}" points to the missing node "{}"'.format(
                    self._format_path(obj), obj.node
                )
            )

        if obj.requires is not None:
            requires = (
                obj.requires
                if isinstance(obj.requires, list)
                else [obj.requires]
            )

            for r in requires:
                if r not in self.ctx.fields_map:
                    self.errors.report(
                        'Link "{}" requires missing field "{}" in the "{}" node'.format(  # noqa: E501
                            obj.name, r, self._format_path()
                        )
                    )

        self._validate_deprecated_duplicates(obj)

    def visit_node(self, obj: Node) -> None:
        node_name = obj.name or "root"
        invalid = [
            f for f in obj.fields if not isinstance(f, self._node_accept_types)
        ]
        if invalid:
            self.errors.report(
                'Node can not contain these types: {} in node "{}"'.format(
                    self._format_types(invalid), node_name
                )
            )
            return

        with self.push_ctx(obj):
            for item in obj.fields:
                self.visit(item)

        duplicates = self._get_duplicates(e.name for e in obj.fields)
        if duplicates:
            self.errors.report(
                'Duplicated names found in the "{}" node: {}'.format(
                    node_name, self._format_names(duplicates)
                )
            )

        if sum((1 for d in obj.directives if isinstance(d, Deprecated))) > 0:
            self.errors.report("Deprecated directive can not be used in Node")

    def visit_root(self, obj: Root) -> None:
        self.visit_node(obj)

    def visit_graph(self, obj: Graph) -> None:
        self.visit_graph_items(obj.items)

    def visit_graph_items(self, items: List[Node]) -> None:
        invalid = [
            f for f in items if not isinstance(f, self._graph_accept_types)
        ]
        if invalid:
            self.errors.report(
                "Graph can not contain these types: {}".format(
                    self._format_types(invalid)
                )
            )
            return

        root = Root(
            list(chain.from_iterable(e.fields for e in items if e.name is None))
        )
        self.visit(root)

        for item in items:
            if item.name is not None:
                self.visit(item)

        duplicates = self._get_duplicates(
            e.name for e in items if e.name is not None
        )
        if duplicates:
            self.errors.report(
                "Duplicated nodes found in the graph: {}".format(
                    self._format_names(duplicates)
                )
            )



from typing import Any, List

from hiku.graph import Graph

from hiku.federation.directive import FieldSet, Key
from hiku.utils import ImmutableDict, to_immutable_dict


def get_keys(graph: Graph, typename: str) -> List[FieldSet]:
    """Get all 'key' directives fields"""
    node = graph.nodes_map[typename]
    return [d.fields for d in node.directives if isinstance(d, Key)]


def representation_to_ident(representation: dict) -> ImmutableDict[str, Any]:
    """Convert representation to ident.
    Ident is a immutable dict that can be used as a key in a dict."""
    return to_immutable_dict(representation, exclude_keys={"__typename"})



from collections import deque

from hiku.denormalize.graphql import DenormalizeGraphQL
from hiku.graph import Graph
from hiku.result import Proxy


class DenormalizeEntityGraphQL(DenormalizeGraphQL):
    def __init__(
        self, graph: Graph, result: Proxy, root_type_name: str
    ) -> None:
        super().__init__(graph, result, root_type_name)
        self._type = deque([graph.__types__[root_type_name]])



from collections import defaultdict
from typing import (
    Optional,
    Dict,
    Awaitable,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

from hiku.cache import CacheInfo, CacheSettings
from hiku.executors.asyncio import AsyncIOExecutor
from hiku.executors.base import SyncAsyncExecutor
from hiku.federation.sdl import print_sdl
from hiku.federation.utils import representation_to_ident
from hiku.engine import (
    BaseEngine,
    InitOptions,
    Query,
    Context,
)
from hiku.executors.queue import Queue
from hiku.federation.version import DEFAULT_FEDERATION_VERSION
from hiku.graph import Graph
from hiku.result import (
    Proxy,
    Index,
    ROOT,
)
from hiku.query import (
    Node,
    Field,
)


if TYPE_CHECKING:
    from hiku.readers.graphql import Operation


V = TypeVar("V")


async def async_result(val: V) -> V:
    return val


class Engine(BaseEngine):
    def __init__(
        self,
        executor: SyncAsyncExecutor,
        cache: Optional[CacheSettings] = None,
        federation_version: int = DEFAULT_FEDERATION_VERSION,
    ) -> None:
        super().__init__(executor, cache)
        if federation_version not in (1, 2):
            raise ValueError("federation_version must be 1 or 2")
        self.federation_version = federation_version

    def execute_service(
        self, graph: Graph, mutation_graph: Optional[Graph]
    ) -> Union[Proxy, Awaitable[Proxy]]:
        idx = Index()
        idx[ROOT.node] = Index()
        idx[ROOT.node][ROOT.ident] = {
            "sdl": print_sdl(
                graph,
                mutation_graph,
                federation_version=self.federation_version,
            )
        }
        result = Proxy(idx, ROOT, Node(fields=[Field("sdl")]))
        if isinstance(self.executor, AsyncIOExecutor):
            return async_result(result)
        return result

    def execute_entities(
        self, graph: Graph, query: Node, ctx: Dict
    ) -> Union[Proxy, Awaitable[Proxy]]:
        path = ("_entities",)
        entities_link = query.fields_map["_entities"]
        query = entities_link.node
        representations = entities_link.options["representations"]

        queue = Queue(self.executor)
        task_set = queue.fork(None)
        query_workflow = Query(queue, task_set, graph, query, Context(ctx))

        type_representations_map = defaultdict(list)

        for rep in representations:
            typename = rep["__typename"]
            ident = representation_to_ident(rep)
            type_representations_map[typename].append(ident)

        for typename in type_representations_map:
            node = graph.nodes_map[typename]
            # TODO(mkind): we probably execute some `resolve_reference`
            #  function and only then run process_node !
            query_workflow.process_node(
                path, node, query, type_representations_map[typename]
            )

        return self.executor.process(queue, query_workflow)

    def execute_query(
        self, graph: Graph, query: Node, ctx: Dict, op: Optional["Operation"]
    ) -> Union[Proxy, Awaitable[Proxy]]:
        query = InitOptions(graph).visit(query)
        queue = Queue(self.executor)
        task_set = queue.fork(None)

        cache = (
            CacheInfo(
                self.cache_settings,
                op.name if op else None,
            )
            if self.cache_settings
            else None
        )

        query_workflow = Query(
            queue, task_set, graph, query, Context(ctx), cache
        )

        query_workflow.start()
        return self.executor.process(queue, query_workflow)

    def execute(
        self,
        graph: Graph,
        query: Node,
        ctx: Optional[Dict] = None,
        op: Optional["Operation"] = None,
    ) -> Union[Proxy, Awaitable[Proxy]]:
        if ctx is None:
            ctx = {}

        if "_entities" in query.fields_map:
            return self.execute_entities(graph, query, ctx)

        return self.execute_query(graph, query, ctx, op)



from typing import (
    Dict,
    Any,
    List,
    Optional,
)

from hiku.graph import Graph
from hiku.query import Node as QueryNode

from hiku.introspection.graphql import GraphQLIntrospection
from hiku.introspection.graphql import AsyncGraphQLIntrospection

from hiku.federation.utils import get_keys


def is_introspection_query(query: QueryNode) -> bool:
    return "__schema" in query.fields_map


def _obj(name: str) -> Dict:
    return {"kind": "OBJECT", "name": name, "ofType": None}


def _non_null(t: Any) -> Dict:
    return {"kind": "NON_NULL", "name": None, "ofType": t}


def _union(name: str, possible_types: Optional[List] = None) -> Dict:
    return {"kind": "UNION", "name": name, "possibleTypes": possible_types}


def _field(name: str, type_: Dict, **kwargs: Any) -> Dict:
    data: Dict = {
        "args": [],
        "deprecationReason": None,
        "description": None,
        "isDeprecated": False,
        "name": name,
        "type": type_,
    }
    data.update(kwargs)
    return data


def _seq_of_nullable(_type: Dict) -> Dict:
    return {
        "kind": "NON_NULL",
        "name": None,
        "ofType": {"kind": "LIST", "name": None, "ofType": _type},
    }


def _seq_of(_type: Dict) -> Dict:
    return _seq_of_nullable({"kind": "NON_NULL", "name": None, "ofType": _type})


def _ival(name: str, type_: Dict, **kwargs: Any) -> Any:
    data = {
        "name": name,
        "type": type_,
        "description": None,
        "defaultValue": None,
    }
    data.update(kwargs)
    return data


def _type(name: str, kind: str, **kwargs: Any) -> Dict:
    data: Dict = {
        "description": None,
        "enumValues": [],
        "fields": [],
        "inputFields": [],
        "interfaces": [],
        "kind": kind,
        "name": name,
        "possibleTypes": [],
    }
    data.update(**kwargs)
    return data


def extend_with_federation(graph: Graph, data: dict) -> None:
    union_types = []
    for node in graph.nodes:
        if get_keys(graph, node.name):
            union_types.append(_obj(node.name))

    data["__schema"]["types"].append(_type("_Any", "SCALAR"))
    data["__schema"]["types"].append(_type("_FieldSet", "SCALAR"))
    data["__schema"]["types"].append(
        _type("_Entity", "UNION", possibleTypes=union_types)
    )
    data["__schema"]["types"].append(
        _type(
            "_Service",
            "OBJECT",
            fields=[_field("sdl", _type("String", "SCALAR"))],
        )
    )

    entities_field = _field(
        "_entities",
        _seq_of_nullable(_union("_Entity", union_types)),
        args=[_ival("representations", _seq_of(_type("_Any", "SCALAR")))],
    )

    service_field = _field(
        "_service",
        _non_null(_obj("_Service")),
    )
    for t in data["__schema"]["types"]:
        if t["kind"] == "OBJECT" and t["name"] == "Query":
            t["fields"].append(entities_field)
            t["fields"].append(service_field)


class BaseFederatedGraphQLIntrospection(GraphQLIntrospection):
    ...


class FederatedGraphQLIntrospection(BaseFederatedGraphQLIntrospection):
    """
    Federation-aware introspection for Federation
    https://www.apollographql.com/docs/federation/federation-spec/#federation-schema-specification
    """

    __directives__ = GraphQLIntrospection.__directives__


class AsyncFederatedGraphQLIntrospection(
    FederatedGraphQLIntrospection, AsyncGraphQLIntrospection
):
    """Adds GraphQL introspection into asynchronous federated graph"""



import dataclasses
from enum import EnumMeta
from typing import Any, List, Optional, TypeVar

EnumType = TypeVar("EnumType", bound=EnumMeta)


@dataclasses.dataclass
class EnumValue:
    name: str
    value: Any


@dataclasses.dataclass
class EnumInfo:
    wrapped_cls: EnumMeta
    name: str
    values: List[EnumValue]


def _process_enum(
    cls: EnumType,
    name: Optional[str] = None,
) -> EnumType:
    if not isinstance(cls, EnumMeta):
        raise TypeError(f"{cls} is not an Enum")

    if not name:
        name = cls.__name__

    values = []
    for item in cls:  # type: ignore
        value = EnumValue(
            item.name,
            item.value,
        )
        values.append(value)

    cls.__enum_info__ = EnumInfo(  # type: ignore
        wrapped_cls=cls,
        name=name,
        values=values,
    )

    return cls


def enum(
    _cls: Optional[EnumType] = None,
    *,
    name: Optional[str] = None,
) -> Any:
    def wrap(cls: EnumType) -> EnumType:
        return _process_enum(cls, name)

    if not _cls:
        return wrap

    return wrap(_cls)



from functools import wraps
from typing import (
    Callable,
    TypeVar,
    List,
    Iterator,
)

from hiku.compat import ParamSpec

from .immutable import ImmutableDict, to_immutable_dict
from .cached_property import cached_property
from .const import const, Const


T = TypeVar("T")
P = ParamSpec("P")


def listify(func: Callable[P, Iterator[T]]) -> Callable[P, List[T]]:
    @wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> List[T]:
        return list(func(*args, **kwargs))

    return wrapper


__all__ = [
    "ImmutableDict",
    "to_immutable_dict",
    "cached_property",
    "const",
    "Const",
]
